@@ -12,9 +12,6 @@
 #define REDIS_CLUSTER_MOD   (REDIS_CLUSTER_SLOTS-1)
 
 /* Complete representation for various commands in RESP */
-#define RESP_MULTI_CMD         "*1\r\n$5\r\nMULTI\r\n"
-#define RESP_EXEC_CMD          "*1\r\n$4\r\nEXEC\r\n"
-#define RESP_DISCARD_CMD       "*1\r\n$7\r\nDISCARD\r\n"
 #define RESP_UNWATCH_CMD       "*1\r\n$7\r\nUNWATCH\r\n"
 #define RESP_CLUSTER_SLOTS_CMD "*2\r\n$7\r\nCLUSTER\r\n$5\r\nSLOTS\r\n"
 #define RESP_ASKING_CMD        "*1\r\n$6\r\nASKING\r\n"
@@ -26,7 +23,7 @@
 /* MOVED/ASK comparison macros */
 #define IS_MOVED(p) (p[0]=='M' && p[1]=='O' && p[2]=='V' && p[3]=='E' && \
                      p[4]=='D' && p[5]==' ')
-#define IS_ASK(p)   (p[0]=='A' && p[1]=='S' && p[3]=='K' && p[4]==' ')
+#define IS_ASK(p)   (p[0]=='A' && p[1]=='S' && p[2]=='K' && p[3]==' ')
 
 /* MOVED/ASK lengths */
 #define MOVED_LEN (sizeof("MOVED ")-1)
@@ -41,15 +38,11 @@
 #define SLOT_STREAM(c,s) (SLOT_SOCK(c,s)->stream)
 #define SLOT_SLAVES(c,s) (c->master[s]->slaves)
 
-/* Macros to access socket and stream for the node we're communicating with */
-#define CMD_SOCK(c) (c->cmd_sock)
-#define CMD_STREAM(c) (c->cmd_sock->stream)
-
-/* Compare redirection slot information with what we have */
-#define CLUSTER_REDIR_CMP(c) \
-    (SLOT_SOCK(c,c->redir_slot)->port != c->redir_port || \
-    ZSTR_LEN(SLOT_SOCK(c,c->redir_slot)->host) != c->redir_host_len || \
-    memcmp(ZSTR_VAL(SLOT_SOCK(c,c->redir_slot)->host),c->redir_host,c->redir_host_len))
+/* Compare redirection slot information with the passed node */
+#define CLUSTER_REDIR_CMP(c, sock) \
+    (sock->port != c->redir_port || \
+    ZSTR_LEN(sock->host) != c->redir_host_len || \
+    memcmp(ZSTR_VAL(sock->host),c->redir_host,c->redir_host_len))
 
 /* Clear out our "last error" */
 #define CLUSTER_CLEAR_ERROR(c) do { \
@@ -62,13 +55,12 @@
 
 /* Protected sending of data down the wire to a RedisSock->stream */
 #define CLUSTER_SEND_PAYLOAD(sock, buf, len) \
-    (sock && !redis_sock_server_open(sock TSRMLS_CC) && sock->stream && !redis_check_eof(sock, 1 TSRMLS_CC) && \
-     php_stream_write(sock->stream, buf, len)==len)
+    (sock && !redis_sock_server_open(sock) && sock->stream && !redis_check_eof(sock, 0, 1) && \
+     redis_sock_write_raw(sock, buf, len) == len)
 
 /* Macro to read our reply type character */
 #define CLUSTER_VALIDATE_REPLY_TYPE(sock, type) \
-    (redis_check_eof(sock, 1 TSRMLS_CC) == 0 && \
-     (php_stream_getc(sock->stream) == type))
+    (redis_check_eof(sock, 1, 1) == 0 && redis_sock_getc(sock) == type)
 
 /* Reset our last single line reply buffer and length */
 #define CLUSTER_CLEAR_REPLY(c) \
@@ -89,11 +81,7 @@
 /* Helper to either return a bool value or add it to MULTI response */
 #define CLUSTER_RETURN_BOOL(c, b) \
     if(CLUSTER_IS_ATOMIC(c)) { \
-        if(b==1) {\
-            RETURN_TRUE; \
-        } else {\
-            RETURN_FALSE; \
-        } \
+        RETURN_BOOL(b); \
     } else { \
         add_next_index_bool(&c->multi_resp, b); \
     }
@@ -128,11 +116,13 @@
     mc->args.len = 0; \
     mc->argc     = 0; \
 
-/* Initialzie a clusterMultiCmd with a keyword and length */
+/* Initialize a clusterMultiCmd with a keyword and length */
 #define CLUSTER_MULTI_INIT(mc, keyword, keyword_len) \
     mc.kw     = keyword; \
     mc.kw_len = keyword_len; \
 
+#define CLUSTER_CACHING_ENABLED() (INI_INT("redis.clusters.cache_slots") == 1)
+
 /* Cluster redirection enum */
 typedef enum CLUSTER_REDIR_TYPE {
     REDIR_NONE,
@@ -141,24 +131,45 @@ typedef enum CLUSTER_REDIR_TYPE {
 } CLUSTER_REDIR_TYPE;
 
 /* MULTI BULK response callback typedef */
-typedef int  (*mbulk_cb)(RedisSock*,zval*,long long, void* TSRMLS_DC);
+typedef int  (*mbulk_cb)(RedisSock*,zval*,long long, void*);
 
-/* Specific destructor to free a cluster object */
-// void redis_destructor_redis_cluster(zend_resource *rsrc TSRMLS_DC);
+/* A list of covered slot ranges */
+typedef struct redisSlotRange {
+    unsigned short low;
+    unsigned short high;
+} redisSlotRange;
+
+/* Simple host/port information for our cache */
+typedef struct redisCachedHost {
+    zend_string *addr;
+    unsigned short port;
+} redisCachedHost;
+
+/* Storage for a cached master node */
+typedef struct redisCachedMaster {
+    redisCachedHost host;
+
+    redisSlotRange *slot;   /* Slots and count */
+    size_t slots;
+
+    redisCachedHost *slave; /* Slaves and their count */
+    size_t slaves;
+} redisCachedMaster;
+
+typedef struct redisCachedCluster {
+    // int rsrc_id;               /* Zend resource ID */
+    zend_string *hash;         /* What we're cached by */
+    redisCachedMaster *master; /* Array of masters */
+    size_t count;              /* Number of masters */
+} redisCachedCluster;
 
 /* A Redis Cluster master node */
 typedef struct redisClusterNode {
-    /* Our Redis socket in question */
-    RedisSock *sock;
-
-    /* A slot where one of these lives */
-    short slot;
-
-    /* Is this a slave node */
-    unsigned short slave;
-
-    /* A HashTable containing any slaves */
-    HashTable *slaves;
+    RedisSock *sock;      /* Our Redis socket in question */
+    short slot;           /* One slot we believe this node serves */
+    zend_llist slots;     /* List of all slots we believe this node serves */
+    unsigned short slave; /* Are we a slave */
+    HashTable *slaves;    /* Hash table of slaves */
 } redisClusterNode;
 
 /* Forward declarations */
@@ -166,16 +177,9 @@ typedef struct clusterFoldItem clusterFoldItem;
 
 /* RedisCluster implementation structure */
 typedef struct redisCluster {
-#if (PHP_MAJOR_VERSION < 7)
-    zend_object std;
-#endif
 
-    /* Timeout and read timeout (for normal operations) */
-    double timeout;
-    double read_timeout;
-
-    /* Are we using persistent connections */
-    int persistent;
+    /* One RedisSock struct for serialization and prefix information */
+    RedisSock *flags;
 
     /* How long in milliseconds should we wait when being bounced around */
     long waitms;
@@ -210,6 +214,11 @@ typedef struct redisCluster {
     /* Flag for when we get a CLUSTERDOWN error */
     short clusterdown;
 
+    /* Key to our persistent list cache and number of redirections we've
+     * received since construction */
+    zend_string *cache_key;
+    uint64_t redirections;
+
     /* The last ERROR we encountered */
     zend_string *err;
 
@@ -220,9 +229,6 @@ typedef struct redisCluster {
     /* The slot where we're subscribed */
     short subscribed_slot;
 
-    /* One RedisSock struct for serialization and prefix information */
-    RedisSock *flags;
-
     /* The first line of our last reply, not including our reply type byte
      * or the trailing \r\n */
     char line_reply[1024];
@@ -238,10 +244,8 @@ typedef struct redisCluster {
     unsigned short     redir_slot;
     unsigned short     redir_port;
 
-#if (PHP_MAJOR_VERSION >= 7)
     /* Zend object handler */
     zend_object std;
-#endif
 } redisCluster;
 
 /* RedisCluster response processing callback */
@@ -313,23 +317,23 @@ typedef struct clusterReply {
     size_t integer;                /* Integer reply */
     long long len;                 /* Length of our string */
     char *str;                     /* String reply */
-    size_t elements;               /* Count of array elements */
+    long long elements;            /* Count of array elements */
     struct clusterReply **element; /* Array elements */
 } clusterReply;
 
 /* Direct variant response handler */
-clusterReply *cluster_read_resp(redisCluster *c, int status_strings TSRMLS_DC);
+clusterReply *cluster_read_resp(redisCluster *c, int status_strings);
 clusterReply *cluster_read_sock_resp(RedisSock *redis_sock,
-    REDIS_REPLY_TYPE type, char *line_reply, size_t reply_len TSRMLS_DC);
+    REDIS_REPLY_TYPE type, char *line_reply, long long reply_len);
 void cluster_free_reply(clusterReply *reply, int free_data);
 
 /* Cluster distribution helpers for WATCH */
-HashTable *cluster_dist_create();
+HashTable *cluster_dist_create(void);
 void cluster_dist_free(HashTable *ht);
 int cluster_dist_add_key(redisCluster *c, HashTable *ht, char *key,
-    strlen_t key_len, clusterKeyVal **kv);
+    size_t key_len, clusterKeyVal **kv);
 void cluster_dist_add_val(redisCluster *c, clusterKeyVal *kv, zval *val
-    TSRMLS_DC);
+   );
 
 /* Aggregation for multi commands like MGET, MSET, and MSETNX */
 void cluster_multi_init(clusterMultiCmd *mc, char *kw, int kw_len);
@@ -340,34 +344,52 @@ void cluster_multi_fini(clusterMultiCmd *mc);
 /* Hash a key to it's slot, using the Redis Cluster hash algorithm */
 unsigned short cluster_hash_key_zval(zval *key);
 unsigned short cluster_hash_key(const char *key, int len);
+unsigned short cluster_hash_key_zstr(zend_string *key);
 
-/* Get the current time in miliseconds */
+/* Validate and sanitize cluster construction args */
+zend_string** cluster_validate_args(double timeout, double read_timeout,
+    HashTable *seeds, uint32_t *nseeds, char **errstr);
+
+void free_seed_array(zend_string **seeds, uint32_t nseeds);
+
+/* Generate a unique hash string from seeds array */
+zend_string *cluster_hash_seeds(zend_string **seeds, uint32_t nseeds);
+
+/* Get the current time in milliseconds */
 long long mstime(void);
 
 PHP_REDIS_API short cluster_send_command(redisCluster *c, short slot, const char *cmd,
-    int cmd_len TSRMLS_DC);
+    int cmd_len);
 
-PHP_REDIS_API void cluster_disconnect(redisCluster *c, int force TSRMLS_DC);
+PHP_REDIS_API void cluster_disconnect(redisCluster *c, int force);
 
-PHP_REDIS_API int cluster_send_exec(redisCluster *c, short slot TSRMLS_DC);
-PHP_REDIS_API int cluster_send_discard(redisCluster *c, short slot TSRMLS_DC);
-PHP_REDIS_API int cluster_abort_exec(redisCluster *c TSRMLS_DC);
-PHP_REDIS_API int cluster_reset_multi(redisCluster *c);
+PHP_REDIS_API int cluster_send_exec(redisCluster *c, short slot);
+PHP_REDIS_API int cluster_send_discard(redisCluster *c, short slot);
+PHP_REDIS_API int cluster_abort_exec(redisCluster *c);
 
 PHP_REDIS_API short cluster_find_slot(redisCluster *c, const char *host,
     unsigned short port);
 PHP_REDIS_API int cluster_send_slot(redisCluster *c, short slot, char *cmd,
-    int cmd_len, REDIS_REPLY_TYPE rtype TSRMLS_DC);
+    int cmd_len, REDIS_REPLY_TYPE rtype);
 
 PHP_REDIS_API redisCluster *cluster_create(double timeout, double read_timeout,
     int failover, int persistent);
-PHP_REDIS_API void cluster_free(redisCluster *c, int free_ctx TSRMLS_DC);
-PHP_REDIS_API int cluster_init_seeds(redisCluster *c, HashTable *ht_seeds);
-PHP_REDIS_API int cluster_map_keyspace(redisCluster *c TSRMLS_DC);
+PHP_REDIS_API void cluster_free(redisCluster *c, int free_ctx);
+PHP_REDIS_API void cluster_init_seeds(redisCluster *c, zend_string **seeds, uint32_t nseeds);
+PHP_REDIS_API int cluster_map_keyspace(redisCluster *c);
 PHP_REDIS_API void cluster_free_node(redisClusterNode *node);
 
-PHP_REDIS_API char **cluster_sock_read_multibulk_reply(RedisSock *redis_sock,
-    int *len TSRMLS_DC);
+/* Functions for interacting with cached slots maps */
+PHP_REDIS_API redisCachedCluster *cluster_cache_create(zend_string *hash, HashTable *nodes);
+PHP_REDIS_API void cluster_cache_free(redisCachedCluster *rcc);
+PHP_REDIS_API void cluster_init_cache(redisCluster *c, redisCachedCluster *rcc);
+
+/* Functions to facilitate cluster slot caching */
+
+PHP_REDIS_API char **cluster_sock_read_multibulk_reply(RedisSock *redis_sock, int *len);
+
+PHP_REDIS_API void cluster_cache_store(zend_string *hash, HashTable *nodes);
+PHP_REDIS_API redisCachedCluster *cluster_cache_load(zend_string *hash);
 
 /*
  * Redis Cluster response handlers.  Our response handlers generally take the
@@ -383,6 +405,26 @@ PHP_REDIS_API void cluster_bool_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster
     void *ctx);
 PHP_REDIS_API void cluster_ping_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
     void *ctx);
+PHP_REDIS_API void cluster_pop_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
+    void *ctx);
+PHP_REDIS_API void cluster_object_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
+    void *ctx);
+PHP_REDIS_API void cluster_lpos_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
+    void *ctx);
+PHP_REDIS_API void cluster_hrandfield_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
+    void *ctx);
+PHP_REDIS_API void cluster_zdiff_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
+    void *ctx);
+PHP_REDIS_API void cluster_zadd_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
+    void *ctx);
+PHP_REDIS_API void cluster_zrandmember_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
+    void *ctx);
+PHP_REDIS_API void cluster_srandmember_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
+    void *ctx);
+PHP_REDIS_API void cluster_set_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
+    void *ctx);
+PHP_REDIS_API void cluster_geosearch_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
+    void *ctx);
 PHP_REDIS_API void cluster_single_line_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
     void *ctx);
 PHP_REDIS_API void cluster_bulk_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
@@ -402,10 +444,15 @@ PHP_REDIS_API void cluster_sub_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *
 PHP_REDIS_API void cluster_unsub_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
     void *ctx);
 
-/* Generic/Variant handler for stuff like EVAL */
+PHP_REDIS_API void cluster_zrange_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
+    void *ctx);
+
 PHP_REDIS_API void cluster_variant_resp(INTERNAL_FUNCTION_PARAMETERS,
     redisCluster *c, void *ctx);
 
+PHP_REDIS_API void cluster_variant_raw_resp(INTERNAL_FUNCTION_PARAMETERS,
+    redisCluster *c, void *ctx);
+
 PHP_REDIS_API void cluster_variant_resp_strings(INTERNAL_FUNCTION_PARAMETERS,
     redisCluster *c, void *ctx);
 
@@ -420,6 +467,8 @@ PHP_REDIS_API void cluster_mbulk_zipstr_resp(INTERNAL_FUNCTION_PARAMETERS,
     redisCluster *c, void *ctx);
 PHP_REDIS_API void cluster_mbulk_zipdbl_resp(INTERNAL_FUNCTION_PARAMETERS,
     redisCluster *c, void *ctx);
+PHP_REDIS_API void cluster_mbulk_dbl_resp(INTERNAL_FUNCTION_PARAMETERS,
+    redisCluster *c, void *ctx);
 PHP_REDIS_API void cluster_mbulk_assoc_resp(INTERNAL_FUNCTION_PARAMETERS,
     redisCluster *c, void *ctx);
 PHP_REDIS_API void cluster_multi_mbulk_resp(INTERNAL_FUNCTION_PARAMETERS,
@@ -439,7 +488,7 @@ PHP_REDIS_API void cluster_msetnx_resp(INTERNAL_FUNCTION_PARAMETERS,
 
 /* Response handler for ZSCAN, SSCAN, and HSCAN */
 PHP_REDIS_API int cluster_scan_resp(INTERNAL_FUNCTION_PARAMETERS,
-    redisCluster *c, REDIS_SCAN_TYPE type, long *it);
+    redisCluster *c, REDIS_SCAN_TYPE type, uint64_t *cursor);
 
 /* INFO response handler */
 PHP_REDIS_API void cluster_info_resp(INTERNAL_FUNCTION_PARAMETERS,
@@ -456,18 +505,29 @@ PHP_REDIS_API void cluster_xrange_resp(INTERNAL_FUNCTION_PARAMETERS,
     redisCluster *c, void *ctx);
 PHP_REDIS_API void cluster_xclaim_resp(INTERNAL_FUNCTION_PARAMETERS,
     redisCluster *c, void *ctx);
+PHP_REDIS_API void cluster_xinfo_resp(INTERNAL_FUNCTION_PARAMETERS,
+    redisCluster *c, void *ctx);
+
+PHP_REDIS_API void cluster_mpop_resp(INTERNAL_FUNCTION_PARAMETERS,
+    redisCluster *c, void *ctx);
+
+/* Custom ACL handlers */
+PHP_REDIS_API void cluster_acl_getuser_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx);
+PHP_REDIS_API void cluster_acl_log_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx);
 
 /* MULTI BULK processing callbacks */
 int mbulk_resp_loop(RedisSock *redis_sock, zval *z_result,
-    long long count, void *ctx TSRMLS_DC);
+    long long count, void *ctx);
 int mbulk_resp_loop_raw(RedisSock *redis_sock, zval *z_result,
-    long long count, void *ctx TSRMLS_DC);
+    long long count, void *ctx);
 int mbulk_resp_loop_zipstr(RedisSock *redis_sock, zval *z_result,
-    long long count, void *ctx TSRMLS_DC);
+    long long count, void *ctx);
+int mbulk_resp_loop_dbl(RedisSock *redis_sock, zval *z_result,
+    long long count, void *ctx);
 int mbulk_resp_loop_zipdbl(RedisSock *redis_sock, zval *z_result,
-    long long count, void *ctx TSRMLS_DC);
+    long long count, void *ctx);
 int mbulk_resp_loop_assoc(RedisSock *redis_sock, zval *z_result,
-    long long count, void *ctx TSRMLS_DC);
+    long long count, void *ctx);
 
 #endif
 
